import sys
import json
import logging

from importlib import reload
from collections.abc import Mapping
from yaml import safe_load, YAMLError

from pathlib import Path
from pygments import highlight
from pygments.formatters.terminal256 import Terminal256Formatter
from pygments.lexers.web import JsonLexer
from pygments.style import Style
from pygments.token import Keyword, Name, String, Number, Token
from tblib import Traceback
from logging import config, LogRecord
from logging_json import JSONFormatter
from logging_json._formatter import (
    _value,
    _extra_attributes,
    default_converter,
)


blank_config = {
    'version': 1,
    'disable_existing_loggers': False,
}


def setup_logger(config_path: str, debug: bool = True):
    """Setup logger."""
    logging.shutdown()
    reload(logging)

    with open(config_path) as _file:
        try:
            log_config = safe_load(_file)
        except (YAMLError, FileNotFoundError) as _exc:
            print(f'Fail load log config: {config_path} - {_exc}')
            log_config = blank_config
        finally:
            config.dictConfig(log_config)
            logging.root.setLevel(logging.DEBUG if debug else logging.ERROR)
            return log_config


class BaseStyle(Style):
    """Base log style."""

    default_style = ""


class StyleDefault(BaseStyle):
    """Log style default or notset."""

    styles = {
        Token:  'bold #808080',
        Keyword: 'bold #808080',
        String: '#c0c0c0',
        Number: '#c0c0c0',
    }


class StyleDebug(BaseStyle):
    """Log style debug."""

    styles = {
        Token: '#5f5faf',
        Name: 'bold #5f5faf',
        String: '#8787ff',
        Number: '#8787ff',
    }


class StyleInfo(BaseStyle):
    """Log style info."""

    styles = {
        Token: '#afd700',
        Name: 'bold #afd700',
        String: '#afaf00',
        Number: '#afaf00',
    }


class StyleWarning(BaseStyle):
    """Log style warning."""

    styles = {
        Token: '#ffaf00',
        Name: 'bold #ffaf00',
        String: '#af8700',
        Number: '#af8700',
    }


class StyleError(BaseStyle):
    """Log style error."""

    styles = {
        Token: '#FF0000',
        Name: 'bold #FF0000',
        String: '#E9967A',
        Number: '#E9967A',
    }


def formatter_from_event_type(event_type: int):
    """Map formatter to event."""
    log_styles = {
        logging.NOTSET: StyleDefault,
        logging.DEBUG: StyleDebug,
        logging.INFO: StyleInfo,
        logging.WARNING: StyleWarning,
        logging.CRITICAL: StyleError,
        logging.ERROR: StyleError,
        logging.FATAL: StyleError,
    }
    try:
        return log_styles[event_type]
    except (AttributeError, KeyError, IndexError):
        return StyleDefault


class PBFormatter(JSONFormatter):
    """PB Log formatter."""

    def format(self, record: LogRecord):
        """Format."""
        super().format(record)

        message = {
            field_name: _value(record, field_value)
            for field_name, field_value in self.fields.items()
        }

        if isinstance(record.msg, Mapping):
            message.update(record.msg)
        else:
            message[self.message_field_name] = super().formatMessage(record)

        message.update(_extra_attributes(record))
        message['msg'] = record.msg or record.message
        if 'color_message' in message:
            message['msg'] = message.pop('color_message')

        if record.name == 'sqlalchemy.engine.Engine':
            _no_sql = ('\n', '\t')
            for _spec in _no_sql:
                message['msg'] = str(message['msg']).replace(_spec, ' ')

        caller = Path(record.pathname)

        message['path'] = '../{parent}/{name}/{script}'.format(
            parent=caller.parent.parent.name,
            name=caller.parent.name,
            script=caller.name,
        )

        if record.exc_info:
            message['exception'] = {
                'type': record.exc_info[0].__name__,
                'msg': str(record.exc_info[1]),
                'stack': Traceback(record.exc_info[2]).as_dict(),
            }

        if len(message) == 1 and self.message_field_name in message:
            return super().formatMessage(record)

        _fmt_style = formatter_from_event_type(record.levelno)

        return highlight(
            json.dumps(
                message,
                default=default_converter,
                indent=4,
                ensure_ascii=False,
            ),
            lexer=JsonLexer(),
            formatter=Terminal256Formatter(style=_fmt_style),
        )


def except_logging(exc_type, exc_value, exc_traceback):
    """Logging exceptions."""
    logging.exception('Uncaught exception', exc_info=exc_value)


sys.excepthook = except_logging
